# File Name: exceptions.py
# Created By: ZW
# Created On: 2022-04-11
# Purpose: provides exception classes and methods for gardnersnake package

# module imports
import sys # for sys.stderr
import json # for printing dicts

# function definitions
#------------------------------------------------------------------------------

# eprint(): printing function to alias print statements to sys.stderr
def eprint(*args, **kw):
    print(*args, file=sys.stderr, **kw)

# edictprint(): printing function to alias print
# statements to sys.stderr, specifically for printing dictionaries
# in parsable formats.
def edictprint(mydict, indent, **kw):
    print(json.dumps(mydict,indent=indent, default=str),
          file=sys.stderr, **kw
          )

# class definitions
#------------------------------------------------------------------------------

# UserError: base class for a custom user error
class UserError(Exception):
    def __init__ (self, message, *args, **kw):
        super(Exception,self).__init__(*args, **kw)
        self.message = message

    def __str__ (self):
        return f"{self.message}"

    def _eprint(self, *args, **kw):
        eprint(*args, **kw)

    def _edictprint(self, mydict, indent, **kw):
        edictprint(mydict, indent, **kw)

    def test_message(self):
        self._eprint(self.__str__())


# ConfigParameterError: raising errors when configuration
# parameters cannot be located in the ConfigurationHelper
# class.
class ConfigParameterError(UserError):
    def __init__ (self, cfg, message, param, *args, **kw):
        super(ConfigParameterError, self).__init__(message, *args, **kw)
        self.cfg = cfg # stores internal representation of config
        self.param = param # parameter that raised the error

    def __str__ (self):
        self._eprint("\nError in the following portion of the configuration..")
        self._edictprint(self.cfg, indent=2)
        return f"Param --> {self.param}. {self.message}"


# RuleConfigurationError: raising errors having to do with
# accessing rule parameters
class ConfigRuleParameterError(ConfigParameterError):
    def __init__ (self, cfg, message, param, rule, *args, **kw):
        super(ConfigRuleParameterError, self).__init__(
            cfg, message, param, *args, **kw
        )
        self.rule = rule

    def __str__ (self):
        self._eprint("\nError in the following portion of the configuration..")
        self._edictprint(self.cfg, indent=2)
        return f"Rule --> {self.rule}, Param --> {self.param}. {self.message}"

#------------------------------------------------------------------------------
#EOF



# File Name: misc.py
# Created By: ZW
# Created On: 2022-04-21
# Purpose: Defines miscillaneous utility functions in gardnersnake
#   until they are rewritten into more comprehensive features

# module imports
#------------------------------------------------------------------------------
from .exceptions import UserError


# function Definitions
#------------------------------------------------------------------------------

# define read_manifest_txt(): takes a text file path pointing to a
# file manifest, reads it in and returns a list of files with
# whitespace stripped. 
# a manifest of file names ()
def read_manifest_txt(filepath):
    with open(filepath, 'r') as manifest:
        files = [line.strip() for line in manifest if len(line.strip()) != 0]
        return files




# File Name: confighelper.py
# Created By: ZW
# Created On: 2022-04-11
# Purpose: a class definition for a configurationHelper that makes
#  snakemake worflows easier to construct

# module imports
from collections import OrderedDict
from .pathutils import get_verified_path
from .exceptions import eprint
from .exceptions import UserError
from .exceptions import ConfigParameterError
from .exceptions import ConfigRuleParameterError

# class definitions
#------------------------------------------------------------------------------

# class ConfigurationHelper: object which
# cleans up much of the code linking run
# conifgurations to snakefile rules.
class ConfigurationHelper:
    def __init__ (self, cfg_dict):
        self.cfg = cfg_dict
        self._validate_config_dict(self.cfg)
        self.globs = self.cfg.copy()
        self.rule_params = self.globs.pop("rule_params")
        self.resource_list = [
            "walltime","nodes",
            "processors_per_node"
        ]

    # define _validate_config_dict() function to ensure minimum requirements
    # for the passed configuration dict --including type checking the dict
    def _validate_config_dict(self,cfg):
        validation_codes = []
        keys_to_check = ["analysis_id", "workdir", "rule_params"]
        outcome = lambda x: validation_codes.append(x)

        # test the type of the passed dict
        outcome(type(cfg) in [dict, OrderedDict])

        # test default required keys
        for k in keys_to_check:
            outcome(k in cfg.keys())

        # verify validation codes
        if not validation_codes[0]: # type check
            eprint(f"passed config: {cfg}")
            raise UserError("Could not validate configuration dictionary")
        elif not all(validation_codes[1:]): # key checks
            msg1 = "Could not Validate Configuration in ConfiguationHelper."
            msg2 = "Please check the listed keys."
            msg = "\n".join([msg1,msg2])
            raise ConfigParameterError(cfg, msg, keys_to_check)
        else:
            return 0

    # define _get_global_param() internal method to return a top level
    # parameter from the passed configuration with optional path handling
    def _get_global_param(self, param, ispath=False,
                         pathtype=None, exists=False):
        # access the parameter, check if we need to do path handling
        # and return accordingly
        try: gparam = self.globs[param]
        except KeyError as ke:
            msg = "Key Error"
            perr = ConfigParameterError(self.globs,msg,param)
            raise perr from ke
        if not ispath: return gparam
        else:
            return get_verified_path(gparam, pathtype, exists)

    # define _get_rule_param() internal method to return a rule-level
    # parameter from the passed configuration with optional path handling
    def _get_rule_param(self, rule, param, ispath=False,
                        pathtype=None, exists=False):
        # access the parameter, check if we need to do path handling
        # and return accordingly.
        try: rparam = self.rule_params[rule][param]
        except KeyError as ke:
            msg = "Key Error"
            rperr = ConfigRuleParameterError(self.rule_params, msg, param, rule)
            raise rperr from ke
        if not ispath: return rparam
        else:
            return get_verified_path(rparam, pathtype, exists)

    # define get_global_param(): exposed wrapper for internal
    # method _get_global_param()
    def get_global_param(self, param, ispath=False, pathtype=None,
                         exists=False, returnPath=False):
        if returnPath:
            return self._get_global_param(param, ispath, pathtype, exists)
        else:
            return str(self._get_global_param(param,
                            ispath, pathtype, exists))

    # define get_rule_param(): exposed wrapper for internal
    # method _get_rule_param()
    def get_rule_param(self, rule, param, ispath=False, pathtype=None,
                       exists=False, returnPath=False):
        if returnPath:
            return self._get_rule_param(rule, param, ispath, pathtype, exists)
        else:
             return str(self._get_rule_param(rule, param,
                                ispath, pathtype, exists))

    # define get_rule_resources(): returns the standard resource set
    # as a dictionary according to the internal resource list-
    # self.resource_list
    def get_rule_resources(self, rule, logdir="logs/", jobid=""):
        # build resource dictionary and return
        rule_spec = self.rule_params[rule]
        addendum = {"log_dir":logdir, "job_id":jobid}
        resources = {}
        resources.update(addendum)

        # build out the rest of the resources through a dictionary comp.
        try:
            rtrigger = None
            for r in self.resource_list:
                rtrigger = r
                resources.update({r:rule_spec[r]})
            return resources
        except KeyError as ke:
            msg = "Could not find resource of the indicated type"
            crpe = ConfigRuleParameterError(rule_spec, msg, rtrigger, rule)
            raise crpe from ke


# EOF



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gardnersnake",
    version="0.1.5",
    author="Zach Weber",
    author_email="zach.weber.813@gmail.com",
    description="Utilities for writing concise snakemake workflows",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zwebbs/gardnersnake",
    project_urls={
        "Bug Tracker": "https://github.com/zwebbs/gardnersnake/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
    ],
    package_dir={"":"src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    entry_points={'console_scripts': {
        'check_directory=scripts.check_directory:main'}
    }
)



# File Name: check_directory.py
# Created By: ZW
# Created On: 2022-04-20
# Purpose: Script to quickly check directory outputs, returning a small file
#   containing a return code (0) if successful, otherwise throwing an error

# module imports
import argparse
from gardnersnake.pathutils import get_verified_path
from gardnersnake.exceptions import UserError

def main():
    # setup CLI for the program
    descr = " validates dynamic directory contents against expectations"
    parser = argparse.ArgumentParser(description=descr, prog="check_directory")
    parser.add_argument("contents", type=str, nargs='+',metavar='FILES',
                        help="set of filepaths to check against dir contents")
    parser.add_argument("directory", type=str, metavar="DIR",
                        help="filepath of directory to verify")
    parser.add_argument("--strict", action="store_true",
                        help="directory should contain only the passed files")
    parser.set_defaults(strict=False)
    parser.add_argument("-o","--output", type=str, metavar="OUT",
                        help="name of return code output file")

    args = parser.parse_args()
    # check the directory contents
    dir_object = get_verified_path(args.directory, "dir", strict=True)
    dir_paths = [f for f in dir_object.glob("**/*") if f.is_file()]
    paths_to_check = [dir_object / f for f in args.contents]

    for filecheck in paths_to_check:
        if filecheck not in dir_paths:
            msg1 = f"\nError. Could not verify contents of {args.directory}"
            msg2 = f"Could not find {filecheck} in {args.directory}"
            raise UserError("\n".join([msg1,msg2]))

    # handle strict directory verification (extra files fails the check)
    if ((args.strict) and (len(paths_to_check) != len(dir_paths))):
        extra_files = dir_paths.copy()
        for f in paths_to_check: extra_files.remove(f)
        msg1 = f"\nError. Could not verify contents of {args.directory}"
        msg2 = f"Found extra files: {extra_files}, with --strict={args.strict}"
        raise UserError("\n".join([msg1,msg2]))

    else: # write out the successful returncode
        if args.output:
            with open(args.output, 'w') as outfile:
                outfile.write('0')
        else:
            msg = "Error. Please pass a file for output"
            raise UserError(msg)










# File Name: __init__.py
# Created By: ZW
# Created On: 2022-04-11
# Purpose: enables module imports from src directory

# specific package imports 
from gardnersnake.confighelper import ConfigurationHelper

# define import all grouping
__all__ = ["confighelper", "exceptions", "pathutils"]




# File Name: pathutils.py
# Created By: ZW
# Created On: 2022-04-11
# Purpose: methods for handling paths in snakemake configs

# module imports
from pathlib import Path
from .exceptions import UserError

# globals
# -----------------------------------------------------------------------------
ACCEPTED_DTYPES = ["dir", "directory", "folder"]
ACCEPTED_FTYPES = ["file", "fp"]
ACCEPTED_TYPES = ["unknown", *ACCEPTED_DTYPES, *ACCEPTED_FTYPES]


# functions
# ------------------------------------------------------------------------------

# define a function to return a verified, expanded path
def get_verified_path(path, pathtype, strict=False):
    pobj = Path(path).expanduser().resolve()  # generate a pathlib Path object
    exists = False  # flag for whether or not the file or directory exists

    # handle the existence check depending on the type of path
    try:
        if pathtype.lower() in ACCEPTED_DTYPES:
            exists = pobj.is_dir()
        elif pathtype.lower() in ACCEPTED_FTYPES:
            exists = pobj.is_file()
        elif pathtype.lower() in ACCEPTED_TYPES:
            exists = pobj.exists()
        else:
            msg1 = f"\nUnknown path type for {path}."
            msg2 = f"\nMust be one of: {ACCEPTED_TYPES}"
            msg3 = f"\nPassed path type (pathtype=): {pathtype}"
            raise UserError("".join([msg1, msg2, msg3]))
    except AttributeError as ae:
        msg4 = "Please pass a string type to argument pathtype"
        raise UserError(msg4) from ae
    # resolve the path object
    if exists and strict:
        return pobj
    elif not strict:
        return pobj
    else:
        msg = f"\nUnable to validate that {path} of type: {pathtype} exists."
        raise UserError(msg)

# EOF
